package ca.yorku.eecs.mack.graffiti;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.pm.ActivityInfo;
import android.content.res.Configuration;
import android.graphics.Point;
import android.media.MediaScannerConnection;
import android.os.Bundle;
import android.os.Environment;
import android.os.Vibrator;
import android.util.Log;
import android.view.Display;
import android.view.LayoutInflater;
import android.view.Surface;
import android.view.View;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.EditText;
import android.widget.ImageView;
import android.widget.TextView;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Random;
import java.util.StringTokenizer;

/**
 * <h1>Graffiti</h1>
 *
 * <h3>Summary</h3>
 *
 * <ul> <li>Android experiment software for text entry using Graffiti or other gesture sets </ul> <p>
 *
 * This Android application expands on Demo_Graffiti. For background information, see the API for Demo_Graffiti as well
 * as the API for the <code>Unistroke</code> class where gesture recognition is performed. This API focuses on features
 * added to Demo_Graffiti, including saving and restoring setup parameters using the <code>SharedPreferences</code>
 * class, storing user performance measurements in data files for follow-up analyses, and auto-correcting input. <p>
 *
 * <h3>Related Reference</h3>
 *
 * The following publication presents research where a variation of this software was used.
 * <p>
 *
 * <ul>
 * <li>
 * <a href="http://www.yorku.ca/mack/chi2012.html">Reducing visual demand for gestural
 * text input on touchscreen devices</a>, by MacKenzie and Castellucci (<i>CHI 2012</i>).
 * <p>
 * </ul>
 * <p>
 *
 *
 * <h3>Related Information</h3>
 *
 * <blockquote> API References: <p>
 *
 * <ul> <li><a href="http://developer.android.com/reference/android/content/SharedPreferences.html">
 * <code>SharedPreferences</code></a>
 * <li><a href="http://developer.android.com/reference/android/content/SharedPreferences.Editor.html">
 * <code>SharedPreferences.Editor</code></a>
 * <li><a href="http://developer.android.com/reference/android/os/Environment.html"> <code>Environment</code></a>
 * </ul> <p>
 *
 * Developer's Guide: <p>
 *
 * <ul> <li><a href="http://developer.android.com/guide/topics/data/data-storage.html">Storage Options</a> (Shared
 * Preferences) </ul> <p>
 *
 * Training: <p>
 *
 * <ul> <li><a href="http://developer.android.com/training/basics/activity-lifecycle/index.html">Managing the Activity
 * Lifecycle</a> <li><a href="http://developer.android.com/training/basics/data-storage/files.html">Saving Files</a>
 * </ul>
 *
 * </blockquote>
 *
 * <!----------------------------------------------------------------------------------><br> <h3>Setup Parameters</h3>
 *
 * The application launches with a setup dialog: <p>
 *
 * <center> <a href="Graffiti-1.jpg"><img src="Graffiti-1.jpg" width="300" alt="image"></a> </center> <p>
 *
 * In all, there are 10 setup parameters, organized in seven spinners and three checkboxes. The setup parameters are
 * explained as follows. <p>
 *
 * <blockquote> <table border="1" cellspacing="0" cellpadding="6" valign="top"> <tr bgcolor="#cccccc"> <th>Parameter
 * <th>Description
 *
 * <tr valign="top"> <td>Participant Code <td>A code to identify the user being tested. In experimental research, a user
 * is typically called a "participant". <p>
 *
 * <tr valign="top"> <td>Session Code <td>A code to identify the session. This code could be used in an a longitudinal
 * experiment, where participants are tested over multiple sessions. Each session might involve an hour of testing, with
 * one session per day over multiple days. <p>
 *
 * Note: There is no setup parameter for "Block code". The block code is generated automatically by the software. <p>
 *
 * <tr valign="top"> <td>Group Code <td>A code to identify the group. If the experiment uses a within-subjects design,
 * participants are usually assigned to groups. This code identifies the group. This procedure is known as
 * <i>counterbalancing</i>, which is used to offset order effects. (See section 5.11 in the course's Suggested Readings
 * for complete details.) <p>
 *
 * <tr valign="top"> <td>Condition Code <td>A code to identify an external condition. This setup item is an editable
 * text field. Tap to popup an input method editor to change the value. <p>
 *
 * The condition code is useful if the experiment included an independent variable pertaining to an environmental
 * condition or a context of use. As an example, an experiment might include "mobility" as an independent variable with
 * levels "sitting" and "walking". The code "Sitting" might be used for the sitting condition and the code "Walking" for
 * the walking condition. These codes appear in the output data to facilitate analyses. BTW, an example of an experiment
 * using this independent variable is described by Fitton et al. in <a href="http://www.yorku.ca/mack/bhci2013
 * .html">Exploring Tilt-Based Text Input For Mobile Devices With Teenagers</a>. <p>
 *
 * <tr valign="top"> <td>Number of Phrases <td>The number of phrases entered for each block of testing. <p>
 *
 * <tr valign="top"> <td>Phrases File <td>A file containing phrases of text that are selected at random and presented to
 * the user for input. The following options appear: <p>
 *
 * <ul> <li>Phrases - a collection of 500 phrases (no punctuation) <li>Phrases100 - a collection of 100 phrases (with
 * punctuation) <li>Quick_Brown_Fox - the phrase "the quick brown fox jumps over the lazy dog" <li>Alphabet - the
 * alphabet ("abcdefghijklmnopqrstuvwxyz") </ul> <p>
 *
 * Note: The actual phrases are in files in <code>/res/values/</code>. <p>
 *
 * <tr valign="top"> <td>Entry mode <td>The gesture set used for input. The following options appear: <p>
 *
 * <ul> <li>Graffiti - the standard Graffiti gesture set <li>Unistrokes - the original Unistrokes gesture set from
 * Goldberg and Richardson's <i>CHI 1993</i> paper, <a href="http://dl.acm.org/citation.cfm?id=169093">Touch Typing With
 * a Stylus</a> <li>Digits - digit gestures <li>Custom - a custom gesture set </ul> <p>
 *
 * See the API for the <a href="Unistroke.html"><code>Unistroke</code></a> class for additional details on the gesture
 * sets. <p>
 *
 * <tr valign="top"> <td>Show Gesture Set <td>A checkbox item. If selected, an image showing the gesture set appears
 * just below the input area. If not selected, the image is hidden. Showing the gesture set is useful for new users who
 * are unfamiliar with the gestures. <p>
 *
 * <tr valign="top"> <td>Lowercase Only <td>A checkbox item. If selected, uppercase characters in the phrases appear as
 * lowercase <p>
 *
 * <tr valign="top"> <td>Show Recognized Text During Entry <td>A checkbox item. If selected, the recognized characters
 * appear in the transcribed text field after each gesture. If not selected, characters do not appear. See "Auto
 * Correct" below for further discussion. </table> </blockquote> <p>
 *
 * The setup parameters may be changed, as appropriate, for the current round of testing. They may be saved by tapping
 * the Save button. This feature was not used in Demo_TiltBall. Having parameters persist from one invocation to the
 * next is useful since it eliminates the need to change parameters each time the application is used. Retrieving and
 * modifying the setup parameters is possible using Android's <code>SharedPreferences</code> and
 * <code>SharedPreferences.Editor</code> classes. Objects of these classes are declared in the
 * <code>GraffitiSetup</code> activity as follows: <p>
 *
 * <pre>
 *      SharedPreferences sp;
 *      SharedPreferences.Editor spe;
 *      </pre>
 *
 * In the activity's <code>onCreate</code> method, a <code>SharedPreferences</code> instance is retrieved as follows:
 * <p>
 *
 * <pre>
 *      sp = this.getPreferences(MODE_PRIVATE);
 * </pre>
 *
 * Setup parameters are retrieved using <i>get</i>-methods. For example, retrieving the default participant code occurs
 * as follows: <p>
 *
 * <pre>
 *      participantCode[0] = sp.getString("participantCode", participantCode[0]);
 * </pre>
 *
 * Participant codes, like the other codes, are stored in arrays. The first entry in the array is the default value. It
 * is over-written with the value associated with the key specified in the first argument of the <code>getString</code>
 * method. This occurs only if such a key-value exists, having been saved earlier. If the value does not exist, then the
 * second argument is returned as a default. As seen above, this amounts leaving the first entry in the array as is.
 * <p>
 *
 * If the Save button is tapped, the current values of the setup parameters are saved via using <i>put</i>-methods in
 * the <code>clickSave</code> callback. Here's the code that saves the participant code: <p>
 *
 * <pre>
 *      spe = sp.edit();
 *      spe.putString("participantCode", participantCode[spinParticipant.getSelectedItemPosition()]);
 * </pre>
 *
 * Of course, the key in <code>putString</code> must be the same as the key in <code>getString</code>. Tapping OK
 * bundles up the setup parameters and launches the main activity, passing the bundle. The process is much the same as
 * in our earlier demo program, Demo_TiltBall. <p>
 *
 * When the program launches, the user is presented with a phrase of text to enter (below left). Data collection begins
 * on the first finger touch on the input panel. Gestures are made and recognized, with the results delivered to a
 * transcribed text field (assuming "Show recognized text during entry" is checked). The image below on the right shows
 * the application in use. <p>
 *
 * <center> <a href="Graffiti-5.jpg"><img src="Graffiti-5.jpg"
 * width="300" alt="image"></a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a href="Graffiti-2.jpg"><img
 * src="Graffiti-2.jpg"
 * width="500" alt="image"></a> </center> <p>
 *
 * <!----------------------------------------------------------------------------------><br> <b>Data Storage</b> <p> As
 * the user enters phrases of text using the selected entry mode (e.g., Graffiti), the software collects data such as
 * the gesture recognition results, timestamps, and gesture shape. At the phrase level, additional measures are
 * calculated such as the entry speed (in words per minute), the character-level error rate (%), and the number of
 * gestures (aka keystrokes) entered per character of text produced (KSPC). <p>
 *
 * As noted above, the Graffiti app includes auto-correct (details below). Two error rates are calculated and
 * saved.
 * One is for the transcribed text as it was produced by the recognizer (i.e., uncorrected). The other is for the
 * corrected text. These data are useful to gauge the performance of the auto-correct algorithm. As an example, the
 * following popup dialog shows the results at the end of a phrase of entry. The error rate for the uncorrected
 * transcribed text was 9.7%. The error rate for the corrected text was 0.0%. (Note: "<code>#</code>" is an unrecognized
 * gesture.) <p>
 *
 * <center> <a href="Graffiti-3.jpg"><img src="Graffiti-3.jpg" width="300" alt="image"></a> </center> <p>
 *
 * The data are saved in files for follow-up analyses. There are three "summary data" files for each block. The file
 * extensions and contents are as follows:
 *
 * <ul> <li>sd1 &ndash; gesture and timestamp data (one line per gesture) <li>sd2 &ndash; phrase summary data (one line
 * per phrase) <li>sd3 &ndash; gesture shape data (a series of <i>x-y</i> coordinates for each gesture) </ul> <p>
 *
 * One of the first tasks in the Graffiti activity is to open and prepare data files for writing. The
 * directory is specified using a final: <p>
 *
 * <pre>
 *     final String WORKING_DIRECTORY = "/GraffitiData/";
 * </pre>
 *
 * In <code>onCreate</code>, the directory path is initialized using <p>
 *
 * <pre>
 *      File dataDirectory = new File(Environment.getExternalStorageDirectory() + WORKING_DIRECTORY);
 * </pre>
 *
 * The <code>getExternalStorageDirectory</code> returns the primary external storage directory. The directory is
 * <i>not</i> external to the device. As noted in the API, <p>
 *
 * <blockquote><blockquote><i> This directory can better be thought of as media/shared storage. It is a filesystem that
 * can hold a relatively large amount of data and that is shared across all applications (does not enforce permissions).
 * Traditionally this is an SD card, but it may also be implemented as built-in storage in a device that is distinct
 * from the protected internal storage and can be mounted as a filesystem on a computer. </i></blockquote></blockquote>
 *
 * <p>
 *
 * Several steps follow. Filenames for the <code>sd1</code>, <code>sd2</code>, and <code>sd3</code> data files are
 * built-up using the setup parameters. "B01" is used within the filename for the block code. Including the setup
 * parameters in the filenames ensures the conditions tested are distinguishable in the names of the output data files.
 * The file system is then queried to see if the files exist. If so, new filenames are built-up using the next block
 * code. This is repeated until the filenames do not conflict with existing files; that is, with files created from
 * previous launches of the application. Then, three buffered writers are opened. With these, we're ready to go. Oh yes,
 * it's important that the manifest file includes the appropriate <code>uses-permission</code> tag:
 *
 * <pre>
 *      &lt;uses-permission android:name="android.permission.WRITE_EXTERNAL_STORAGE" /&gt;
 *      &lt;uses-permission android:name="android.permission.READ_EXTERNAL_STORAGE" /&gt;
 * </pre>
 *
 * As entry proceeds, data are collected. At the end of a block, data are written to the <code>sd1</code>,
 * <code>sd2</code>, and <code>sd3</code> files using the <code>write</code> method of the <code>BufferedWriter</code>
 * class. This is followed by <code>flush</code> and <code>close</code>. Of course, all this occurs in
 * <code>try-catch</code> statements. <p>
 *
 * There is one additional small detail. If the Android device is connected to a host machine via a USB cable, then the
 * following statement is needed to ensure the data files are viewable in the host machine's file manager (e.g., Windows
 * Explorer): <p>
 *
 * <pre>
 *      MediaScannerConnection.scanFile(this, new String[] { f1.getAbsolutePath(),
 *           f2.getAbsolutePath(), f3.getAbsolutePath() }, null, null);
 * </pre>
 *
 * Here are examples of the data files: <p>
 *
 * <ul> <li>sd1 &ndash; <a href="Graffiti-P01-S01-B01-G01-C01-Graffiti.sd1">
 * <code>Graffiti-P01-S01-B01-G01-C01-Graffiti.sd1</code></a>
 *
 * <li>sd2 &ndash; <a
 * href="Graffiti-P01-S01-B01-G01-C01-Graffiti.sd2">
 * <code>Graffiti-P01-S01-B01-G01-C01-Graffiti.sd2</code></a>
 *
 * <li>sd3 &ndash; <a href="Graffiti-P01-S01-B01-G01-C01-Graffiti.sd3">
 * <code>Graffiti-P01-S01-B01-G01-C01-Graffiti.sd3</code></a>
 * </ul> <p>
 *
 * The sd2 file is comma-delimited to facilitate importing into a spreadsheet or statistics application. It includes a
 * header line that identifies the data in each column. Each row contains data for the conditions as well as for the the
 * human performance measures gathered for that phrase of input. Here is a look at the data in the sd2 file above, as
 * they appear in Microsoft Excel: <p>
 *
 * <center> <a href="Graffiti-4.jpg"><img src="Graffiti-4.jpg" width="1000" alt="image"></a> </center> <p>
 *
 * <!----------------------------------------------------------------------------------><br> <b>Auto Correct</b> <p>
 *
 * The <code>Graffiti</code> class includes an auto-correct algorithm. The motivation for this follows from a
 * simple observation: Users enter gestures more quickly if they do <i>not</i> monitor the results of recognition after
 * each gesture. Of course, users monitor the recognition result for a reason &ndash; to see if a gesture was correctly
 * recognized. Wouldn't it be nice if users didn't need to monitor their progress &ndash; if an error occurs, it gets
 * automatically corrected! That's a lofty goal. Achieving this goal requires an auto-correct algorithm. <p>
 *
 * Of course, if the recognition result does not appear after each gesture, there is nothing to monitor. In this case,
 * users are likely to proceed more expeditiously. This, combined with a good auto-correct algorithm, has the potential
 * to yield input that is both fast and accurate. <p>
 *
 * Of course, errors will occur. There are three possibilities: (a) a wrong gesture, (b) an ill-formed and misrecognized
 * gesture, or (c) an unrecognized gesture. Auto correct is intended to work in concert with an input modality where
 * users do not monitor their input on a gesture-by-gesture basis. This is the purpose of the "Show recognized text
 * during entry" setup parameter (see above). The rationale and goals are further elaborated in MacKenzie and
 * Castellucci's <a href="http://www.yorku.ca/mack/chi2012.html">Reducing Visual Demand for Gestural Text Input on
 * Touchscreen Devices</a>. <p>
 *
 * The initial design of the auto-correct algorithm is described in Tinwala and MacKenzie's <a
 * href="http://www.yorku.ca/mack/nordichi2010.html">Eyes-free Text Entry with Error Correction on Touchscreen Mobile
 * Devices</a>. Of course, the complete details are in the source code. Here's a high-level description of the code.
 * Auto correct is handled by the <code>SpellCheck</code> class. In <code>Graffiti</code>, a
 * <code>SpellCheck</code> object is declared, instantiated, and initialized as follows: <p>
 *
 * <pre>
 *      SpellCheck sc;
 *      ...
 *      sc = new SpellCheck();
 *      ...
 *      sc.setDictionary(wf);
 * </pre>
 *
 * The <code>wf</code> object passed to <code>setDictionary</code> is a word-frequency array &ndash; a dictionary
 * &ndash; for the language of input. The raw word+frequency values for the dictionary are stored in
 * <code>/res/values/wordfreq.xml</code>. <p> At the end of each phrase of input (in the <code>doEndOfPhrase</code>
 * method in <code>Graffiti</code>), the uncorrected text phrase is passed to <code>SpellCheck</code>'s
 * <code>correctPhrase</code> method. A corrected phrase is returned: <p>
 *
 * <pre>
 *      String transcribedPhraseCorrected = sc.correctPhrase(transcribedBuffer.toString());
 * </pre>
 *
 * The actual correction occurs in <code>correctPhrase</code> using a look-up and compare technique applied to each word
 * in the phrase. If a word is in the dictionary, it is assumed correct. If a word is not in the dictionary, a
 * minimum-string-distance algorithm is used to compare the erroneous word with words of a similar size in the
 * dictionary. The closest and highest-frequency dictionary word is substituted for the erroneous word. There's a bit
 * more to it, of course. Consult the <code>correctPhrase</code> method in <code>SpellCheck</code> for the complete
 * details. <p>
 *
 * <b>Note:</b> The auto-correct algorithm is only used if the "Phrases file" option in the setup dialog is set to
 * "phrases2.txt". As well, case is ignored (e.g., "Jedi" and "jedi" are considered the same). <p>
 *
 * @author (c) Scott MacKenzie, 2011-2022
 */
public class GraffitiActivity extends Activity implements GraffitiPanel.OnStrokeListener
{
    final static String MYDEBUG = "MYDEBUG"; // for Log.i messages

    final String WORKING_DIRECTORY = "/GraffitiData/";
    final String APP = "Graffiti";
    final String SD2_HEADER = "App,Participant,Session,Block,Group,Condition,Entry_mode,Keystrokes,Characters," +
            "Time_(s),Speed_(wpm),TranscribedTextErrorRate(%),CorrectedTextErrorRate(%),KSPC\n";
    final String GRAFFITI = "Graffiti";
    final String UNISTROKES = "Unistrokes";
    final String DIGITS = "Digits";
    final int PULSE_DURATION = 15;

    // setup parameters (that need to be member variables)
    int numberOfPhrases;
    boolean lowercaseOnly, showRecognizedTextDuringEntry;
    // ...end setup parameters

    //Point[] stroke;
    int strokeCount; // number of strokes in a phrase
    String[] wordFreqTmp;
    SpellCheck spellCheck;
    boolean spellCheckOn = false; // enable spell check for phrases file only
    int phraseCount;
    boolean done = false;
    boolean endOfPhrase, firstGestureInPhrase;
    StringBuilder transcribedBuffer;
    String presentedBuffer;

    GraffitiPanel graffitiPanel;
    ImageView imagePanel;
    TextView presentedText;
    EditText transcribedText;
    Random r = new Random();
    ArrayList<Sample> samples;
    long t1, elapsedTimeForPhrase;
    long timeStartOfPhrase;
    String[] phrases;
    BufferedWriter sd1, sd2, sd3;
    File f1, f2, f3;
    String sd2Leader; // sd2Leader to identify conditions for data written to sd2 files.
    Vibrator vib;
    int defaultOrientation;

    // compute typing speed in wpm given text entered and time in ms
    public static float wpm(String text, long msTime)
    {
        float speed = text.length();
        speed = speed / (msTime / 1000.0f) * (60 / 5);
        return speed;
    }

    // Called when the activity is first created
    @Override
    public void onCreate(Bundle savedInstanceState)
    {
        super.onCreate(savedInstanceState);

        // init study parameters from shared preferences passed from setup dialog
        Bundle b = getIntent().getExtras();
        String participantCode = b.getString("participantCode");
        String sessionCode = b.getString("sessionCode");
        String groupCode = b.getString("groupCode");
        String conditionCode = b.getString("conditionCode");
        numberOfPhrases = b.getInt("numberOfPhrases");
        String phrasesFile = b.getString("phrasesFile");
        String entryMode = b.getString("entryMode");
        boolean showGestureSet = b.getBoolean("showGestureSet");
        lowercaseOnly = b.getBoolean("lowercaseOnly");
        showRecognizedTextDuringEntry = b.getBoolean("showRecognized");

        // determine which UI view to use (i.e., with or without the gestures showing)
        if (!showGestureSet || entryMode.equals(DIGITS))
            setContentView(R.layout.main_hide_gestures);
        else
            setContentView(R.layout.main);

        // get this device's default orientation
        defaultOrientation = getDefaultDeviceOrientation();

        // force the app to appear in the device's default orientation (and stay that way)
        if (defaultOrientation == Configuration.ORIENTATION_LANDSCAPE)
            setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_LANDSCAPE);
        else
            setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT);

        // get references to UI components
        presentedText = (TextView)findViewById(R.id.presented);
        transcribedText = (EditText)findViewById(R.id.transcribed);
        graffitiPanel = (GraffitiPanel)findViewById(R.id.paintPanel);
        imagePanel = (ImageView)findViewById(R.id.imgGestureSet);

        // set the Graffiti panel's width and height, according to the display size
        WindowManager wm = (WindowManager)this.getSystemService(Context.WINDOW_SERVICE);
        Display display = wm.getDefaultDisplay();
        int displayWidth = display.getWidth();
        int displayHeight = display.getHeight();

        graffitiPanel.width = displayWidth - 20;
        graffitiPanel.height = displayHeight / 2;

        vib = (Vibrator)getSystemService(Context.VIBRATOR_SERVICE);

        // load the phrases file (based on setup choice)
        if (phrasesFile.equals("phrases2"))
        {
            // only enable spell checking if input uses the standard phrases file.
            spellCheckOn = true;
            phrases = getResources().getStringArray(R.array.phrases2);

        } else if (phrasesFile.equals("phrases100"))
            phrases = getResources().getStringArray(R.array.phrases100);
        else if (phrasesFile.equals("quickbrownfox"))
            phrases = getResources().getStringArray(R.array.quickbrownfox);
        else if (phrasesFile.equals("alphabet"))
            phrases = getResources().getStringArray(R.array.alphabet);

        // list for graffiti gestures on the main touch panel
        graffitiPanel.setOnStrokeListener(this);

        // set the entry mode (Graffiti, Unistrokes, or Digits)
        if (entryMode.equals(GRAFFITI))
            graffitiPanel.setDictionary(Unistroke.GRAFFITI);
        else if (entryMode.equals(UNISTROKES))
            graffitiPanel.setDictionary(Unistroke.UNISTROKES);
        else if (entryMode.equals(DIGITS))
            graffitiPanel.setDictionary(Unistroke.DIGITS);

        // setup the dictionary for spell checking (i.e., auto correct)
        spellCheck = new SpellCheck();
        wordFreqTmp = getResources().getStringArray(R.array.wordFreq);
        SpellCheck.WordFreq[] wf = new SpellCheck.WordFreq[wordFreqTmp.length];
        StringTokenizer st;
        for (int i = 0; i < wf.length; ++i)
        {
            st = new StringTokenizer(wordFreqTmp[i]);
            String word = st.nextToken();
            int freq = Integer.parseInt(st.nextToken());
            wf[i] = spellCheck.getWordFreq(word, freq);
        }
        spellCheck.setDictionary(wf); // done!

        samples = new ArrayList<Sample>();
        transcribedBuffer = new StringBuilder();

        // ===================
        // File initialization
        // ===================

        // make a working directory (if necessary) to store data files
        File dataDirectory = new File(Environment.getExternalStorageDirectory() + WORKING_DIRECTORY);
        if (!dataDirectory.exists() && !dataDirectory.mkdirs())
        {
            Log.e(MYDEBUG, "ERROR --> FAILED TO CREATE DIRECTORY: " + WORKING_DIRECTORY);
            super.onDestroy(); // cleanup
            this.finish(); // terminate
        }
        Log.i(MYDEBUG, "Working directory=" + dataDirectory);

        /*
         * The following do-loop creates data files for output and a string sd2Leader to write to the sd2
         * output files.  Both the filenames and the sd2Leader are constructed by combining the setup parameters
         * so that the filenames and sd2Leader are unique and also reveal the conditions used for the block of input.
         *
         * The block code begins "B01" and is incremented on each loop iteration until an available
         * filename is found.  The goal, of course, is to ensure data files are not inadvertently overwritten.
         */
        int blockNumber = 0;
        do
        {
            ++blockNumber;
            String blockCode = String.format(Locale.CANADA, "B%02d", blockNumber);
            String baseFilename = String.format("%s-%s-%s-%s-%s-%s-%s", APP, participantCode,
                    sessionCode, blockCode, groupCode, conditionCode, entryMode);
            f1 = new File(dataDirectory, baseFilename + ".sd1");
            f2 = new File(dataDirectory, baseFilename + ".sd2");
            f3 = new File(dataDirectory, baseFilename + ".sd3");

            // also make a comma-delimited leader that will begin each data line written to the sd2 file
            sd2Leader = String.format("%s,%s,%s,%s,%s,%s,%s", APP, participantCode, sessionCode,
                    blockCode, groupCode, conditionCode, entryMode);
        } while (f1.exists() || f2.exists());

        try
        {
            sd1 = new BufferedWriter(new FileWriter(f1));
            sd2 = new BufferedWriter(new FileWriter(f2));
            sd3 = new BufferedWriter(new FileWriter(f3));

            // output header in sd2 file
            sd2.write(SD2_HEADER, 0, SD2_HEADER.length());
            sd2.flush();
        } catch (IOException e)
        {
            Log.e(MYDEBUG, "ERROR OPENING DATA FILES! e=" + e.toString());
            super.onDestroy();
            this.finish();
        } // end file initialization

        // initialize Graffiti-related stuff
        strokeCount = 0;

        // ============================
        // CONFIGURE STROKE IMAGE PANEL
        // ============================

        if (showGestureSet && !entryMode.equals(DIGITS))
        {
            if (entryMode.equals(GRAFFITI))
                imagePanel.setImageResource(R.drawable.graffiti);
            else if (entryMode.equals(UNISTROKES))
                imagePanel.setImageResource(R.drawable.unistrokes);

            imagePanel.setAdjustViewBounds(true);
            imagePanel.setMaxWidth(displayWidth);
        }

        transcribedText.requestFocus();

        // prevent soft keyboard from popping up
        getWindow().setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_ALWAYS_HIDDEN);

        phraseCount = 0;
        doNewPhrase();

    } // end onCreate

    // get the default orientation of the device
    public int getDefaultDeviceOrientation()
    {
        WindowManager windowManager = (WindowManager)getSystemService(WINDOW_SERVICE);
        Configuration config = getResources().getConfiguration();
        int rotation = windowManager.getDefaultDisplay().getRotation();

        if (((rotation == Surface.ROTATION_0 || rotation == Surface.ROTATION_180) && config.orientation ==
                Configuration.ORIENTATION_LANDSCAPE)
                || ((rotation == Surface.ROTATION_90 || rotation == Surface.ROTATION_270) && config.orientation ==
                Configuration.ORIENTATION_PORTRAIT))
            return Configuration.ORIENTATION_LANDSCAPE;
        else
            return Configuration.ORIENTATION_PORTRAIT;
    }

    // executed on finger lift (callback from GraffitiPanel)
    public void onStroke(GraffitiEvent ge)
    {
        if (done)
        {
            doDone();
            return;
        }

        String raw = ge.raw;
        int charCode = ge.charCode;
        int type = ge.type;
        Point[] p = ge.gesture;
        long timeStamp = ge.timeStampFingerUp;

        ++strokeCount;
        StringBuilder sb = new StringBuilder();
        sb.append(String.format("Trace_data_for_%s\n", raw));
        for (Point point : p)
            sb.append(String.format(Locale.CANADA, "%d,%d\n", point.x, point.y));
        sb.append("END_OF_STROKE\n");

        // write trace data to sd3 file
        try
        {
            sd3.write(sb.toString(), 0, sb.length());
            sd3.flush();
        } catch (IOException e)
        {
            Log.e(MYDEBUG, "ERROR WRITING TO sd3 TRACE DATA FILE!\n" + e);
            this.finish();
        }

        switch (type)
        {
            case GraffitiEvent.TYPE_UNRECOGNIZED:
                transcribedBuffer.append(Character.toString((char)GraffitiEvent.CHAR_UNRECOGNIZED));
                vib.vibrate(PULSE_DURATION);
                break;

            case GraffitiEvent.TYPE_BACKSPACE:
                if (transcribedBuffer.length() > 0)
                    transcribedBuffer.deleteCharAt(transcribedBuffer.length() - 1);
                break;

            case GraffitiEvent.TYPE_ENTER:
                endOfPhrase = true; // ENTER stroke not added to samples
                break;

            case GraffitiEvent.TYPE_SPACE:
            case GraffitiEvent.TYPE_ALPHA:
            case GraffitiEvent.TYPE_NUMERIC:
            case GraffitiEvent.TYPE_SYMBOL:
            case GraffitiEvent.TYPE_TAP:
                transcribedBuffer.append(Character.toString((char)charCode));
                break;

            case GraffitiEvent.TYPE_NONE:
            case GraffitiEvent.TYPE_NORTH:
            case GraffitiEvent.TYPE_NORTH_EAST:
            case GraffitiEvent.TYPE_SOUTH_EAST:
            case GraffitiEvent.TYPE_NORTH_WEST:
                break; // likely a state change, nothing to do

            default: // should never reach here
                Log.e(MYDEBUG, "ERROR. UNKNOWN CHARACTER! raw=" + raw);
                super.onDestroy();
                this.finish();
                break;
        }

        if (endOfPhrase)
            doEndOfPhrase();
        else
            doEndOfStroke(transcribedBuffer, raw, timeStamp);
    }

    private void doEndOfStroke(StringBuilder sb, String rawArg, long timeStampArg)
    {
        if (showRecognizedTextDuringEntry)
            transcribedText.setText(sb);

        // move cursor to end
        transcribedText.setSelection(transcribedText.getText().length());

        graffitiPanel.clear();

        if (firstGestureInPhrase)
        {
            timeStartOfPhrase = timeStampArg;
            firstGestureInPhrase = false;
        }
        elapsedTimeForPhrase = timeStampArg - timeStartOfPhrase;
        samples.add(new Sample(elapsedTimeForPhrase, rawArg));
    }

    public void doNewPhrase()
    {
        String phrase = phrases[r.nextInt(phrases.length)];
        presentedBuffer = lowercaseOnly ? phrase.toLowerCase(Locale.CANADA) : phrase;
        presentedText.setText(presentedBuffer);
        transcribedBuffer.setLength(0);
        transcribedText.setText(transcribedBuffer);

        // write presented text phrase to sd3 file
        try
        {
            sd3.write(presentedBuffer + "\n", 0, presentedBuffer.length() + 1);
            sd3.flush();
        } catch (IOException e)
        {
            Log.e(MYDEBUG, "ERROR WRITING TO sd3 TRACE DATA FILE!\n" + e);
            super.onDestroy();
            this.finish();
        }

        t1 = 0;
        strokeCount = 0;
        samples.clear();
        endOfPhrase = false;
        firstGestureInPhrase = true;
    }

    public void doEndOfPhrase()
    {
        if (presentedBuffer == null)
            return; // "Enter" button pressed before anything entered

        String presentedPhrase = presentedBuffer.toLowerCase(Locale.getDefault()).trim();
        String transcribedPhraseUncorrected = transcribedBuffer.toString().toLowerCase(Locale.getDefault()).trim();

        String transcribedPhraseCorrected = spellCheckOn ? spellCheck.correctPhrase(transcribedPhraseUncorrected)
                : transcribedPhraseUncorrected;

        StringBuilder resultsString = new StringBuilder("Thank you!\n\n");
        resultsString.append(String.format("Presented...\n   %s\n", presentedPhrase));
        resultsString.append(String.format("Transcribed...\n   %s\n", transcribedPhraseUncorrected));

        if (spellCheckOn)
            resultsString.append(String.format("Corrected...\n   %s\n", transcribedPhraseCorrected));

        StringBuilder sd2Data = new StringBuilder(100);
        sd2Data.append(String.format("%s,", sd2Leader));

        // output number of strokes (aka keystrokes)
        sd2Data.append(String.format(Locale.CANADA, "%d,", strokeCount));

        // output number of characters entered
        sd2Data.append(String.format(Locale.CANADA, "%d,", transcribedPhraseUncorrected.length()));

        // output time in seconds
        float d = elapsedTimeForPhrase / 1000.0f;
        sd2Data.append(String.format(Locale.CANADA, "%.2f,", d));

        // output speed in words per minute
        d = wpm(transcribedPhraseUncorrected, elapsedTimeForPhrase);
        resultsString.append(String.format(Locale.CANADA, "Entry speed: %.2f wpm\n", d));
        sd2Data.append(String.format(Locale.CANADA, "%.2f,", d));

        // output error rate for transcribed text
        // MSD2 s1s2 = new MSD2(presentedPhrase, transcribedPhraseUncorrected);
        MSD s1s2 = new MSD(presentedPhrase, transcribedPhraseUncorrected);
        d = (float)s1s2.getErrorRateNew();
        resultsString.append(String.format(Locale.CANADA, "Error rate (transcribed): %.2f%%\n", d));
        sd2Data.append(String.format(Locale.CANADA, "%.2f,", d));

        // output error rate for corrected text
        if (spellCheckOn)
        {
            // MSD2 s1s3 = new MSD2(presentedPhrase, transcribedPhraseCorrected);
            MSD s1s3 = new MSD(presentedPhrase, transcribedPhraseCorrected);
            d = (float)s1s3.getErrorRateNew();
            resultsString.append(String.format(Locale.CANADA, "Error rate (corrected): %.2f%%\n", d));
            sd2Data.append(String.format(Locale.CANADA, "%f,", d));
        } else
            sd2Data.append("na,"); // not applicable

        // output KSPC (keystrokes per character)
        d = (float)strokeCount / transcribedPhraseUncorrected.length();
        resultsString.append(String.format(Locale.CANADA, "KSPC: %.4f\n\n", d));
        resultsString.append("Click OK to continue");
        sd2Data.append(String.format(Locale.CANADA, "%.4f\n", d)); // end of line too!

        // dump data
        StringBuilder sd1Stuff = new StringBuilder();
        sd1Stuff.append(String.format("%s\n", presentedPhrase));
        sd1Stuff.append(String.format("%s\n", transcribedPhraseUncorrected));
        sd1Stuff.append(String.format("%s\n", transcribedPhraseCorrected));

        for (Sample s : samples)
            sd1Stuff.append(String.format("%s\n", s.toString()));
        sd1Stuff.append("-----\n");

        // write to data files
        try
        {
            sd1.write(sd1Stuff.toString(), 0, sd1Stuff.length());
            sd1.flush();
            sd2.write(sd2Data.toString(), 0, sd2Data.length());
            sd2.flush();
            String s2s3 = transcribedPhraseUncorrected + "\n" + transcribedPhraseCorrected + "\n-----\n";
            sd3.write(s2s3, 0, s2s3.length());
            sd3.flush();
        } catch (IOException e)
        {
            Log.e("MYDEBUG", "ERROR WRITING TO DATA FILE!\n" + e);
            super.onDestroy();
            this.finish();
        }

        // present results to user
        showResultsDialog(resultsString.toString());

        // check if last phrase in block
        ++phraseCount;
        if (phraseCount < numberOfPhrases)
            doNewPhrase();
        else
            done = true; // will terminate on next stroke/tap (allows the user to see results from last phrase)
    }

    private void showResultsDialog(String text)
    {
        LayoutInflater inflater = (LayoutInflater)this.getSystemService(LAYOUT_INFLATER_SERVICE);
        View layout = inflater.inflate(R.layout.results_dialog, (ViewGroup)findViewById(R.id.results_layout));

        // Set text
        TextView results = (TextView)layout.findViewById(R.id.resultsArea);
        results.setText(text);

        // Initialize the dialog
        AlertDialog.Builder parameters = new AlertDialog.Builder(this);
        parameters.setView(layout).setCancelable(false).setNeutralButton("OK", new DialogInterface.OnClickListener()
        {
            public void onClick(DialogInterface dialog, int id)
            {
                dialog.cancel(); // close this dialog
            }
        }).show();
    }

    private void doDone()
    {
        try
        {
            sd1.close();
            sd2.close();
            sd3.close();

			/*
             * Make the saved data files visible in Windows Explorer. There seems to be bug doing
			 * this with Android 4.4. I'm using the following code, instead of sendBroadcast. See...
			 * 
			 * http://code.google.com/p/android/issues/detail?id=38282
			 */
            MediaScannerConnection.scanFile(this, new String[] {f1.getAbsolutePath(), f2.getAbsolutePath(),
                    f3.getAbsolutePath()}, null, null);

        } catch (IOException e)
        {
            Log.e(MYDEBUG, "ERROR CLOSING DATA FILES! e=" + e);
        }
        super.onDestroy();
        this.finish();
    }

    // -------------------------------------------------------
    // Sample - simple class to hold a timestamp and keystroke
    // -------------------------------------------------------
    private class Sample
    {
        private long time;
        private String key;

        Sample(long timeArg, String keyArg)
        {
            time = timeArg;
            key = keyArg;
        }

        public String toString()
        {
            return time + ", " + key;
        }
    }
}